/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  MECCA -- the Maximus Embedded Command Compiler (Advanced)              *
 *                                                                         *
 *  This is the source module for the Maximus Embedded Command             *
 *  Compiler (Advanced), a superset of OECC.  MECCA is 100%                *
 *  backwards-compatible with OECC, right down to the actual               *
 *  bytes output.  MECCA is a "one-and-a-half-pass" compiler,              *
 *  meaning that it makes only one pass over the source file.              *
 *  MECCA will compile as it goes, and build tables of forward             *
 *  references as it goes along.  If it turns out that there               *
 *  were forward references, then MECCA will a second pass over            *
 *  the object code (ONLY if necessary), and update any labels             *
 *  found therein.  In this way, MECCA is as fast as a one-pass            *
 *  compiler when forward refs are NOT used, but still faster              *
 *  than a two-pass compiler when they ARE used.                           *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#define MECCA
#define SILT
#define INITIALIZE
#define NOVARS
#define MAX_NOPROTO

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "prog.h"
#include "alc.h"
#include "max.h"
#include "mecca.h"

struct _symbol *symbol[MAX_SYM]={NULL};
unsigned char tokenbuf[MAX_TOKENLEN+1];

int lcol,
    scol,
    n_scol,
    max_len,
    define,
    errs=0;

int cdecl main(int argc,char *argv[])
{
#ifdef OS_2   /* OS/2's version string causes this to wrap over 80 cols */
  Hello("MECCA","Maximus Embedded Command Compiler(Advanced)",VERSION,"1990");
#else
  Hello("MECCA","The Maximus Embedded Command Compiler (Advanced)",VERSION,"1990");
#endif

  if (argc < 2)
  {
    printf("Usage: MECCA <infile> [outfile]\n\n");

    printf("<infile> uses a default extension of .MEC.  <infile> can also specify\n\n");

    printf("If <outfile> is not specified, then MECCA will default to the\n");
    printf("input filename, with an extension of .BBS.  If <infile> is specified using\n"
           "wildcards, then <outfile> should specify the PATH in which to place\n"
           "the compiled output files.  If no <outfile> is specified when using\n"
           "wildcards, then the output files will be placed in the same directory\n"
           "as the input files.\n");

    return 1;
  }

  Init_Table();

  Parse_Args(argc,argv);

  if (errs)
    printf("\n*** %d errors in compile ***\n",errs);
  else printf("\nDone!\n");
  return 0;
}


void Parse_Args(int argc,char *argv[])
{
  FFIND *ff;

  char inname[PATHLEN],
       outname[PATHLEN],
       av1[PATHLEN],
       av2[PATHLEN];

  char temp[PATHLEN],
       *p;

  int found_extension,
      time_comp,
      ret,
      x;

  *av1='\0';
  *av2='\0';
  time_comp=FALSE;

  for (x=1;x < argc;x++)
  {
    if (*argv[x]=='-')
      time_comp=TRUE;
    else if (*av1)
      strcpy(av2,argv[x]);
    else strcpy(av1,argv[x]);
  }

  if ((ff=FindOpen(av1,0))==NULL)
  {
    strcpy(temp,av1);
    strcat(temp,".MEC");

    if ((ff=FindOpen(temp,0))==NULL)
    {
      errs++;
      printf("File not found: `%s'\n",av1);
      return;
    }
  }

  for (ret=0;ret==0;ret=FindNext(ff))
  {
    strcpy(inname,av1);
    max_len=12;

    if ((p=strrchr(inname,'\\')) != NULL ||
        (p=strrchr(inname,':')) != NULL)
    {
      *(p+1)='\0';
      max_len=strlen(inname)+12;
      strcpy(p+1,ff->ff_name);
    }
    else strcpy(inname,ff->ff_name);

    if ((p=strchr(av1,'.')) != NULL && *(p+1)=='\0')
      strcat(inname,".");

    if (strchr(av1,'*') || strchr(av1,'?'))
    {
      if (*av2)
      {
        strcpy(outname,av2);

        if (outname[strlen(outname)-1] != '\\' && outname[strlen(outname)-1] != ':')
          strcat(outname,"\\");
      }
      else strcpy(outname,inname);

      if ((p=strrchr(outname,'\\')) != NULL ||
          (p=strrchr(outname,':')) != NULL)
      {
        *(p+1)='\0';
        strcat(outname,ff->ff_name);

        if ((p=strrchr(outname,'.')) != NULL)
          *p='\0';
      }
      else mknl(outname);
    }
    else if (argc >= 3) /* No wildcards */
    {
      strcpy(outname,av2);

      if (outname[strlen(outname)-1]=='\\' || outname[strlen(outname)-1]==':')
      {
        strcat(outname,ff->ff_name);

        if ((p=strrchr(outname,'.')) != NULL)
          *p='\0';
      }
    }
    else mknl(outname);

    found_extension=FALSE;
    p=inname;

    if (strpbrk(p,PATHDELIM) != NULL) /* Find if there's a filename extension */
    {
      while ((p=strpbrk(p,PATHDELIM)) != NULL)
      {
        if (*p=='\\' || *p=='/')
          found_extension=FALSE;

        if (*p=='.')
          found_extension=TRUE;

        p++;
      }
    }

    strcpy(temp,inname);

    if (found_extension)
    {
      if (isnl(outname))
      {
        (*strchr(temp,'.'))='\0';   /* Chop off extension */
        strcpy(outname,temp);
      }
    }
    else
    {
      if (isnl(outname))
        strcpy(outname,inname);

      strcat(inname,".MEC");
    }

    found_extension=FALSE;
    p=outname;

    if (strpbrk(p,PATHDELIM) != NULL) /* Find if there's a filename extension */
    {
      while ((p=strpbrk(p,PATHDELIM)) != NULL)
      {
        if (*p=='\\' || *p=='/')
          found_extension=FALSE;

        if (*p=='.')
          found_extension=TRUE;

        p++;
      }
    }

    if (! found_extension)
      strcat(outname,".BBS");

    lcol=7;
    n_scol=0;

    Compile(inname,outname,0,time_comp);
  }

  FindClose(ff);
}


/* mode==0: only used on first call.  mode==1: used on second and
   consecutive calls, means we open in semi-APPEND mode, and don't
   update the label table after this input file.                    */

void pascal Compile(char *inname,char *outname,int mode,int time_comp)
{
  union stamp_combo in_st,
                    out_st;

  char *amode;

  FILE *infile,
       *outfile;

  FFIND *inf,
        *outf;

  int ch,
      tokenidx,
      lastch,
      lastcount,
      x,y;

  lastch=-1;
  lastcount=0;

  if (!mode)
  {
    printf("Compiling %-*s to %-*s: ",max_len,strupr(inname),max_len,strupr(outname));

    if (eqstri(inname,outname))
    {
      errs++;
      printf("Can't compile to itself!  Skipping.\n");
      return;
    }
  }

  if (time_comp)
  {
    if ((inf=FindOpen(inname,0)) != NULL)
    {
      if ((outf=FindOpen(outname,0)) != NULL)
      {
        in_st.dos_st.date=inf->ff_fdate;
        in_st.dos_st.time=inf->ff_ftime;

        out_st.dos_st.date=outf->ff_fdate;
        out_st.dos_st.time=outf->ff_ftime;

        FindClose(outf);

        if (!GEdate(&in_st,&out_st))
        {
          printf("Skipped due to file date.\n");
          return;
        }
      }

      FindClose(inf);
    }
  }

  if ((infile=fopen(inname,"r"))==NULL)
  {
    if (mode==1)
      printf("    ");

    printf("Error opening `%s' for input!\n",inname);
    errs++;

    return;
  }

  if (mode==0)
    amode="w+b";
  else amode="r+b";

  if ((outfile=fopen(outname,amode))==NULL)
  {
    fclose(infile);

    printf("Error opening `%s' for output!\n",outname);
    errs++;

    return;
  }

  if (mode==0)
    printf("\n");

  if (mode==1)
    fseek(outfile,0L,SEEK_END);

  if (setvbuf(infile,NULL,_IOFBF,INOUT_BUFFER))
    NoMem();

  if (setvbuf(outfile,NULL,_IOFBF,INOUT_BUFFER))
    NoMem();

  while (1)
  {
    if ((ch=getc(infile))==EOF)  /* End of file */
      break;

    if (ch==lastch && ch != '[')
      lastcount++;
    else Compress_Sequence();

    if (ch=='\n')
    {
      putc('\r',outfile);
      putc('\n',outfile);
      lastcount=0;
      lastch=-1;
    }
    else if (ch=='[')
    {
      ch=-1;
      tokenidx=0;
      tokenbuf[0]='\0';

      while (ch != ']' && tokenbuf[0] != '[' && tokenidx < MAX_TOKENLEN) /* Read in the token */
      {
        ch=getc(infile);
        tokenbuf[tokenidx++]=(unsigned char)ch;
      }

      if (tokenbuf[0] != '[')
      {
        lastch=-1;
        lastcount=0;

        tokenbuf[tokenidx-1]='\0';

        if (tokenidx==MAX_TOKENLEN)
        {
          printf("    Token too long: `%s'\n",tokenbuf);
          errs++;
        }

        tokenbuf[tokenidx]='\0';

        strlwr(tokenbuf);

        Parse_Token(outname,&outfile);
      }
    }
  }

  if (mode==0)
  {
    for (x=0;symbol[x];x++)
    {
      if (symbol[x]->location==-1)
      {
        printf("    Undefined symbol: `%s'\n",symbol[x]->name);
        errs++;
      }
      else
      {
        for (y=0;y < symbol[x]->offsetnum;y++)
        {
          fseek(outfile,symbol[x]->offset[y],SEEK_SET);
          fprintf(outfile,"%05ld",symbol[x]->location);
        }
      }

      free(symbol[x]);
      symbol[x]=NULL;
    }
  }

  Compress_Sequence();

  fclose(infile);
  fclose(outfile);

  return;
}




void pascal Parse_Token(char *outname,FILE **outfile)
{
  struct _inf inf;
  
  char *p;

  int verbnum;

  p=strtok(tokenbuf,TOKENDELIM);

  while (p)
  {
    if (*p=='/')
    {
      p++;
      define=TRUE;
    }
    else define=FALSE;

    if ((verbnum=sssearch(p,verbs,verb_table_size))==-1)
    {
      if (isdigitstr(p))            /* A "[65]"-type ASCII code */
        fputc(atoi(p),*outfile);
      else Process_Label(p,define,*outfile); /* Else it's undefined, maybe a label? */
    }
    else if (verbs[verbnum].fptr)
    {
      /* Copy any pertinent info into struct */

      inf.p=&p;
      inf.infile=NULL;
      inf.outfile=outfile;
      inf.outname=outname;
      inf.verbnum=verbnum;

      /* And call the handling routine */

      (*verbs[verbnum].fptr)(&inf);
    }
    else
    {
      if (define)
      {
        printf("\a    Error!  Reserved word used as a token: `%s'!\n",p);
        errs++;
      }
      else Output_Code(verbnum,*outfile,&lcol);
    }

    if (p != NULL)
      p=strtok(NULL,TOKENDELIM);
  }
}



void pascal P_On(struct _inf *inf)
{
  FILE *outfile;
  char *p;
  
  int ch,
      x;
    
  outfile=*inf->outfile;
    
  if (*inf->p && ftell(outfile) != 0)  /* Get the next colour token */
  {
    p=strtok(NULL,TOKENDELIM);

    fseek(outfile,-1L,SEEK_CUR);
    ch=getc(outfile);
    fseek(outfile,-1L,SEEK_CUR);

    x=Colour_Num(p);

    if (x >= 8)
      x -= 8;

    ch += x*16;

    lcol=ch;

    if (ch >= 160)
    {
      fseek(outfile,-1L,SEEK_CUR);
      ch -= 128;
    }

    putc(ch,outfile);
  }
}

void pascal P_Comment(struct _inf *inf)
{
  *inf->p=NULL;
}

void pascal P_Include(struct _inf *inf)
{
  char *p;
  
  p=strtok(NULL,TOKENDELIM);

  if (p)
  {
    fclose(*inf->outfile);

    Compile(p,inf->outname,1,FALSE);

    if ((*inf->outfile=fopen(inf->outname,"r+b"))==NULL)
    {
      printf("\a    Error reopening `%s' for output!\n",*inf->outname);
      errs++;
      return;
    }

    if (setvbuf(*inf->outfile,NULL,_IOFBF,INOUT_BUFFER))
      NoMem();
  
    fseek(*inf->outfile,0L,SEEK_END);
  }
}

void pascal P_Copy(struct _inf *inf)
{
  FILE *copyfile,
       *outfile;
     
  char *p;
  int ch;
  
  p=strtok(NULL,TOKENDELIM);
  outfile=*inf->outfile;

  if (p && (copyfile=fopen(p,"rb")) != NULL)
  {
    while (1)
    {
      ch=getc(copyfile);

      if (! feof(copyfile))
        putc(ch,outfile);
      else break;
    }

    fclose(copyfile);
  }
  else
  {
    printf("    Error opening `%s'!\n",p);
    errs++;
  }
}
      
 
void pascal P_Setpriv(struct _inf *inf)
{
  char *p;
  
  p=strtok(NULL,TOKENDELIM);

  if (p)
    fprintf(*inf->outfile,"\x17""s%c",toupper(*p));
}

void pascal P_Menu_Cmd(struct _inf *inf)
{
  int x;

  char *p,
       temp[PATHLEN];
  
  p=strtok(NULL,TOKENDELIM);

  if (! isdigit(*p))
  {
    for (x=0;x < silt_table_size;x++)
      if (eqstri(silt_table[x].token,p))
        break;

    if (x==silt_table_size)
    {
      printf("\a    Invalid menu command: `%s'!\n",p);
      x=0;
    }
    else sprintf(p=temp,"%d",(int)silt_table[x].opt);
  }

  if (p)
    fprintf(*inf->outfile,"\x17""r%c",(unsigned char)atoi(p));
}



void pascal P_Save(struct _inf *inf)
{
  (void)inf;
  scol=lcol;
}

void pascal P_Load(struct _inf *inf)
{
  if (n_scol)
    n_scol--;

  if (scol != -1)
    Put_Colour(*inf->outfile,scol);

  lcol=scol;
}

void pascal P_Fg(struct _inf *inf)
{
  char *p;
  int x;
  
  p=strtok(NULL,TOKENDELIM);

  if (p)
  {
    x=Colour_Num(p);
    lcol=(lcol & 0xf0)+x;
    Put_Colour(*inf->outfile,lcol);
  }
}

void pascal P_Bg(struct _inf *inf)
{
  char *p;
  int x;
  
  p=strtok(NULL,TOKENDELIM);

  if (p)
  {
    x=Colour_Num(p);
    lcol=(lcol & 0x8f)+(x << 4);
    Put_Colour(*inf->outfile,lcol);
  }
}

void pascal P_Steady(struct _inf *inf)
{
  Put_Colour(*inf->outfile,lcol);
}

void pascal P_Dim(struct _inf *inf)
{
  Put_Colour(*inf->outfile,lcol &= ~0x08);
}

void pascal P_Bright(struct _inf *inf)
{
  Put_Colour(*inf->outfile,lcol |= 0x08);
}

void pascal P_Label(struct _inf *inf)
{
  char *p;

  p=strtok(NULL,TOKENDELIM);

  if (p)
    Process_Label(p,TRUE,*inf->outfile);
}

void pascal P_Restore(struct _inf *inf)
{
  (void)inf;
  lcol=scol;
}

void pascal P_Priv(struct _inf *inf)
{
  FILE *outfile;
  
  int verbnum,
      type,
      x;
  
  char priv_letter,
       keys,
       ch;
  
  char *p,
       *p2,
       *s;
     
  p=strtok(NULL,TOKENDELIM);
  p2=strtok(NULL,TOKENDELIM);

  outfile=*inf->outfile;
  verbnum=inf->verbnum;
  
  if (p)
  {
    keys=0;
    
    if ((s=strchr(p,'/')) != NULL)
      while (*++s)
        if (*s >= '1' && *s <= '8')
          keys |= 1 << (*s-'1');
    
    priv_letter=toupper(*p);

    type=Determine_Type(p2);

    if (type==1)
      Invert_Logical(&verbnum);

    if (verbnum==num_gt || verbnum==num_above)
      fprintf(outfile,"\x10""L%c\x10""X%c",priv_letter,priv_letter);
    else if (verbnum==num_lt || verbnum==num_below)
    {
      if (type==2)
        fprintf(outfile,"\x10""F%c",priv_letter);
      else fprintf(outfile,"\x10""B%c\x10""X%c",priv_letter,priv_letter);
    }
    else if (verbnum==num_eq || verbnum==num_equal)
      fprintf(outfile,"\x10""Q%c",priv_letter);
    else if (verbnum==num_ne || verbnum==num_unequal ||
             verbnum==num_notequal)
      fprintf(outfile,"\x10""X%c",priv_letter);
    else if (verbnum==num_ge || verbnum==num_ae)
      fprintf(outfile,"\x10""L%c",priv_letter);
    else if (verbnum==num_le || verbnum==num_be)
      fprintf(outfile,"\x10""B%c",priv_letter);

    if (type==2 && verbnum != num_lt && verbnum != num_below)
      fprintf(outfile,"\x0f""Q\r\n");
    
    if (type==1)
      ch='X';
    else ch='Q';

    for (x=0;x <= 7;x++)
      if (keys & (1 << x))
        fprintf(outfile,"\x10%c%c",ch,'1'+(char)x);

  }
}

void pascal P_Iftime(struct _inf *inf)
{
  int hh,mm;
  
  char temp[PATHLEN],
       *p,
       *p2;

  int x;
         
  p=strtok(NULL,TOKENDELIM);
  p2=strtok(NULL,TOKENDELIM);

  if ((x=sssearch(p,verbs,verb_table_size))==-1 ||
      !IsValidTimeVerb(x))
  {
    printf("\a    Invalid time token: `%s'!\n",p);
    errs++;
  }
  else
  {
    if (sscanf(p2,"%d:%d",&hh,&mm) != 2)
    {
      printf("\a    Invalid time format: `%s'!\n",p2);
      errs++;
    }
    else
    {
      strcpy(temp,"\x17t");

      if (x==num_gt || x==num_above)
        strcat(temp,"\x01");
      else if (x==num_lt || x==num_below)
        strcat(temp,"\x02");
      else if (x==num_eq || x==num_equal)
        strcat(temp,"\x03");
      else if (x==num_ne || x==num_unequal || x==num_notequal)
        strcat(temp,"\x04");
      else if (x==num_ge || x==num_ae)
        strcat(temp,"\x05");
      else if (x==num_le || x==num_be)
        strcat(temp,"\x06");

      strcat(temp,"%c%c");
      fprintf(*inf->outfile,temp,hh+1,mm+1);
    }
  }
}


void pascal Process_Label(char *p,int define,FILE *outfile)
{
  int x;

  for (x=0;symbol[x] && x < MAX_SYM;x++)
    if (eqstr(symbol[x]->name,p))
      break;

  if (x==MAX_SYM)
  {
    printf("\a    Error!  Too many symbols defined!\n");
    errs++;
  }
  else
  {
    if (symbol[x]==NULL) /* Allocate new token */
    {
      if ((symbol[x]=(struct _symbol *)malloc(sizeof(struct _symbol)))==NULL)
        NoMem();

      strcpy(symbol[x]->name,p);
      symbol[x]->offsetnum=0;
      symbol[x]->location=-1;
    }

    if (define)
      symbol[x]->location=ftell(outfile);
    else
    {
      if (symbol[x]->location==-1)
        symbol[x]->offset[symbol[x]->offsetnum++]=ftell(outfile);

      if (symbol[x]->offsetnum==MAX_OFFSETS)
      {
        printf("\a    Too many references, label `%s'\n",p);
        errs++;
      }
      else if (!define)
      {
        if (symbol[x]->location != -1)
          fprintf(outfile,"%05ld",symbol[x]->location);
        else fputs("?????",outfile);
      }
    }
  }
}

void pascal P_Goto(struct _inf *inf)
{
  fputs("\x0fV",*inf->outfile);
}

void pascal P_Repeat(struct _inf *inf)
{
  Output_Code(inf->verbnum,*inf->outfile,&lcol);
}

void pascal P_Repeatseq(struct _inf *inf)
{
  Output_Code(inf->verbnum,*inf->outfile,&lcol);
}


void pascal P_Locate(struct _inf *inf)
{
  Output_Code(inf->verbnum,*inf->outfile,&lcol);
}

/*
int cdecl vbcmp(struct _table *v1,struct _table *v2)
{
  return (strcmp(v1->verb,v2->verb));
}
*/


void NoMem(void)
{
  printf("    Ran out of memory!\n");
  exit(1);
}


int sssearch(char *key,struct _table base[],unsigned int num)
{
  int x,
      lastx=-1,
      lasthi,
      lastlo;

  char *s,*t;

  lasthi=num;
  lastlo=0;

  while (1)
  {
    x=((lasthi-lastlo) >> 1)+lastlo;

    if (lastx==x)
      return -1;

    lastx=x;

    for (s=key,t=base[x].verb;*s==*t;s++,t++)
      if (! *s)
        return (x);                       /* Found a match */

    if (*s > *t)
      lastlo=x;
    else
      lasthi=x;
  }

  return -1;
}

int pascal GEdate(union stamp_combo *s1,union stamp_combo *s2)
{
  /* Compare the date of `s1' and `s2'.  Definately the long way around    *
   * the barn, but portable.                                               */

  if (s1->msg_st.date.yr >= s2->msg_st.date.yr)
  {
    if (s1->msg_st.date.yr > s2->msg_st.date.yr)
      return TRUE;
    else if (s1->msg_st.date.mo >= s2->msg_st.date.mo)
    {
      if (s1->msg_st.date.mo > s2->msg_st.date.mo)
        return TRUE;
      else if (s1->msg_st.date.da >= s2->msg_st.date.da)
      {
        if (s1->msg_st.date.da > s2->msg_st.date.da)
          return TRUE;
        else if (s1->msg_st.time.hh >= s2->msg_st.time.hh)
        {
          if (s1->msg_st.time.hh > s2->msg_st.time.hh)
            return TRUE;
          else if (s1->msg_st.time.mm >= s2->msg_st.time.mm)
          {
            if (s1->msg_st.time.mm > s2->msg_st.time.mm)
              return TRUE;
            else if (s1->msg_st.time.ss >= s2->msg_st.time.ss)
              return TRUE;
          }
        }
      }
    }
  }

  return FALSE;
}


void Put_Colour(FILE *outfile,int col)
{
  if (col >= 160)
    fprintf(outfile,"\x16\x01%c",(char)(col-128));
  else fprintf(outfile,"\x16\x01\x10%c",(char)(col | 0x80));
}


int Colour_Num(char *name)
{
  int vn;
  
  vn=sssearch(name,verbs,verb_table_size);

  if (vn==-1)
  {
    printf("    Invalid colour: `%s'\n",name);
    errs++;

    return 7;
  }
  else return (verbs[vn].colour);
}

int isdigitstr(char *s)
{
  while (*s)
    if (! isdigit(*s++))
      return FALSE;

  return TRUE;
}




void pascal Invert_Logical(int *verbnum)
{
  if (*verbnum==num_gt || *verbnum==num_above)
    *verbnum=num_le;
  else if (*verbnum==num_lt || *verbnum==num_below)
    *verbnum=num_ge;
  else if (*verbnum==num_eq || *verbnum==num_equal)
    *verbnum=num_ne;
  else if (*verbnum==num_ne || *verbnum==num_unequal ||
           *verbnum==num_notequal)
    *verbnum=num_eq;
  else if (*verbnum==num_ge || *verbnum==num_ae)
    *verbnum=num_lt;
  else if (*verbnum==num_le || *verbnum==num_be)
    *verbnum=num_gt;
}




int pascal IsValidTimeVerb(int verbnum)
{
  if (verbs[verbnum].fptr==P_Priv)
    return TRUE;
  else return FALSE;
}


int Determine_Type(char *s)
{
  if (!s || eqstri(s,"see") ||  eqstri(s,"show") || eqstri(s,"line"))
    return 0;
  else if (eqstri(s,"skip") || eqstri(s,"noshow"))
    return 1;
  else if (eqstri(s,"quit") || eqstri(s,"exit"))
    return 2;

  printf("\a    Invalid action: `%s'\n",s);

  return 0;
}



/* Output a translation code to the *.BBS file, and set the colour token    *
 * (if required.)                                                           */

void pascal Output_Code(int verbnum,FILE *outfile,int *lcol)
{
  if (verbnum < verb_table_size)
  {
    fputs(verbs[verbnum].translation,outfile);

    if (verbs[verbnum].colour != -1)
      *lcol=verbs[verbnum].colour;
  }
}


